"""The Bjerksund and Stensland (1993)"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..european.generalised_black_scholes import price as bs_price

norm = NormalDist()
cdf = norm.cdf

def _phi(
        S: float,
        T: float,
        gamma_: float,
        h: float,
        i: float,
        r: float,
        b: float,
        v: float,
) -> float:
    lambda_ = (-r + gamma_ * b + 0.5 * gamma_ * (gamma_ - 1) * v ** 2) * T
    d = -(log(S / h) + (b + (gamma_ - 0.5) * v ** 2) * T) / (v * sqrt(T))
    kappa = 2 * b / v ** 2 + 2 * gamma_ - 1
    return (
        exp(lambda_) * S ** gamma_ * (
            cdf(d) - (i / S) ** kappa * cdf(d - 2 * log(i / S) / (v * sqrt(T)))
        )
    )


def _call_price(
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:

    if b >= r:
        # We can use Black-Scholes as it is never optimal to exercise before
        # maturity.
        return bs_price(True, S, K, T, r, b, v)

    beta = (
        (1 / 2 - b / v ** 2)
        + sqrt((b / v ** 2 - 1 / 2) ** 2 + 2 * r / v ** 2)
    )
    b_infinity = beta / (beta - 1) * K
    b0 = max(K, r / (r - b) * K)
    ht = -(b * T + 2 * v * sqrt(T)) * b0 / (b_infinity - b0)
    i = b0 + (b_infinity - b0) * (1 - exp(ht))
    alpha = (i - K) * i ** (-beta)
    if S >= i:
        return S - K
    else:
        return (
            alpha * S ** beta
            - alpha * _phi(S, T, beta, i, i, r, b, v)
            + _phi(S, T, 1, i, i, r, b, v)
            - _phi(S, T, 1, K, i, r, b, v)
            - K * _phi(S, T, 0, i, i, r, b, v)
            + K * _phi(S, T, 0, K, i, r, b, v)
        )


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    """The Bjerksund and Stensland (1993) American approximation.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.

    Returns:
        float: The price of the option.
    """
    if is_call:
        return _call_price(S, K, T, r, b, v)
    else:
        # Use the Bjerksund and Stensland put-call transformation
        return _call_price(K, S, T, r - b, -b, v)



"""European binomial"""

from math import comb, exp, log, sqrt

def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> float:
    """A European binomial option pricing tree.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        float: The price of the option.
    """
        
    dt = T / n
    u = exp(v * sqrt(dt))
    d = 1 / u
    a = exp(b * dt)
    p = (a - d) / (u - d)
    A = int(log(K / (S * d ** n)) / log(u / d)) + 1

    sum = 0
    if is_call:
        for j in range(A, n+1):
            sum += comb(n, j) * p ** j * (1 - p) ** (n - j) * (
                S * u ** j * d ** (n - j) - K
            )
    else:
        for j in range(A):
            sum += comb(n, j) * p ** j * (1 - p) ** (n - j) * (
                K - S * u ** j * d ** (n - j)
            )

    return exp(-r * T) * sum



"""Leisen Reimer"""

from math import exp, log, nan, sqrt
from typing import Literal, Tuple, Union

def _sign(n: Union[float, int]) -> Literal[-1, 0, 1]:
    if n > 0:
        return 1
    elif n < 0:
        return -1
    else:
        return 0

def _odd(n: int) -> int:
    s = _sign(n)
    n = abs(n)
    if n % 2 == 0:
        n += 1
    return n * s
 
def greeks(
        is_european: bool,
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
        n: int
) -> Tuple[float, float, float, float]:
    """Leisen-Reimer binomial tree.

    Args:
        is_european (bool): Tue for European, false for American.
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.
        n (int): The number of the steps in the tree.

    Returns:
        Tuple[float, float, float, float]: The price, delta, gamma, theta.
    """

    n = _odd(n)
    z = 1 if is_call else -1
    
    d1 = (log(S / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    # Using Preizer-Pratt inversion method 2
    hd1 = 0.5 + _sign(d1) * (
        0.25
        - 0.25 * exp(-(d1 / (n + 1 / 3 + 0.1 / (n + 1))) ** 2 * (n + 1 / 6))
    ) ** 0.5
    hd2 = 0.5 + _sign(d2) * (
        0.25
        - 0.25 * exp(-(d2 / (n + 1 / 3 + 0.1 / (n + 1))) ** 2 * (n + 1 / 6))
    ) ** 0.5
    
    dT = T / n
    p = hd2
    u = exp(b * dT) * hd1 / hd2
    d = (exp(b * dT) - p * u) / (1 - p)
    df = exp(-r * dT)
    option_value = [
        max(0, z * (S * u ** i * d ** (n - i) - K))
        for i in range(n+1)
    ]
    
    delta = gamma = theta = nan

    for j in range(n-1, -1, -1):
        for i in range(j+1):
            if is_european:
                option_value[i] = (
                    p * option_value[i + 1]
                    + (1 - p) * option_value[i]
                ) * df
            else:
                option_value[i] = max(
                    (z * (S * u ** i * d ** (j - i) - K)),
                    (
                        p * option_value[i + 1]
                        + (1 - p) * option_value[i]
                    ) * df
                )

        if j == 2:
            gamma = (
                (option_value[2] - option_value[1]) / (S * u ** 2 - S * u * d)
                - (option_value[1] - option_value[0]) / (S * u * d - S * d ** 2)
            ) / (0.5 * (S * u ** 2 - S * d ** 2))
            theta = option_value[1]
            
        if j == 1:
            delta = (option_value[1] - option_value[0]) / (S * u - S * d)

    theta = (theta - option_value[0]) / (2 * dT) / 365

    return option_value[0], delta, gamma, theta



"""Barone-Adesi and Whaley"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..european.generalised_black_scholes import price as bs_price

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf
inv_cdf = norm.inv_cdf


def _kc(
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    """Newton Raphson algorithm to solve for the critical commodity price for a
    Call.

    Args:
        K (float): The strike.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        b (float): The asset growth.
        v (float): The volatility.

    Returns:
        float: The price.
    """

    # Calculate the seed value Si
    n = 2 * b / v ** 2
    m = 2 * r / v ** 2
    q2u = (-(n - 1) + sqrt((n - 1) ** 2 + 4 * m)) / 2
    su = K / (1 - 1 / q2u)
    h2 = -(b * T + 2 * v * sqrt(T)) * K / (su - K)
    Si = K + (su - K) * (1 - exp(h2))

    k = 2 * r / (v ** 2 * (1 - exp(-r * T)))
    d1 = (log(Si / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
    q2 = (-(n - 1) + sqrt((n - 1) ** 2 + 4 * k)) / 2
    lhs = Si - K
    rhs = (
        bs_price(True, Si, K, T, r, b, v) +
        (1 - exp((b - r) * T) * cdf(d1)) * Si / q2
    )
    bi = (
        exp((b - r) * T) * cdf(d1) * (1 - 1 / q2) +
        (1 - exp((b - r) * T) * cdf(d1) / (v * sqrt(T))) / q2
    )
    epsilon = 0.000001
    # Using the Newton Raphson algorithm solve for Si
    while abs(lhs - rhs) / K > epsilon:
        Si = (K + rhs - bi * Si) / (1 - bi)
        d1 = (log(Si / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
        lhs = Si - K
        rhs = (
            bs_price(True, Si, K, T, r, b, v) +
            (1 - exp((b - r) * T) * cdf(d1)) * Si / q2
        )
        bi = (
            exp((b - r) * T) * cdf(d1) * (1 - 1 / q2) +
            (1 - exp((b - r) * T) * pdf(d1) / (v * sqrt(T))) / q2
        )

    return Si


def _call_price(
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:

    if b >= r:
        return bs_price(True, S, K, T, r, b, v)

    Sk = _kc(K, T, r, b, v)
    n = 2 * b / v ** 2
    k = 2 * r / (v ** 2 * (1 - exp(-r * T)))
    d1 = (log(Sk / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
    q2 = (-(n - 1) + sqrt((n - 1) ** 2 + 4 * k)) / 2
    a2 = (Sk / q2) * (1 - exp((b - r) * T) * cdf(d1))
    if S < Sk:
        return (
            bs_price(True, S, K, T, r, b, v)
            + a2 * (S / Sk) ** q2
        )
    else:
        return S - K


def _kp(
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    # Newton Raphson algorithm to solve for the critical commodity price for a Put

    # Calculation of seed value, Si
    n = 2 * b / v ** 2
    m = 2 * r / v ** 2
    q1u = (-(n - 1) - sqrt((n - 1) ** 2 + 4 * m)) / 2
    su = K / (1 - 1 / q1u)
    h1 = (b * T - 2 * v * sqrt(T)) * K / (K - su)
    Si = su + (K - su) * exp(h1)

    k = 2 * r / (v * 2 * (1 - exp(-r * T)))
    d1 = (log(Si / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
    q1 = (-(n - 1) - sqrt((n - 1) ** 2 + 4 * k)) / 2
    lhs = K - Si
    rhs = (
        bs_price(False, Si, K, T, r, b, v)
        - (1 - exp((b - r) * T) * cdf(-d1)) * Si / q1
    )
    bi = (
        -exp((b - r) * T) * cdf(-d1) * (1 - 1 / q1)
        - (1 + exp((b - r) * T) * pdf(-d1) / (v * sqrt(T))) / q1
    )
    epsilon = 0.000001
    # Using the Newton Raphson algorithm, solve for Si.
    while abs(lhs - rhs) / K > epsilon:
        Si = (K - rhs + bi * Si) / (1 + bi)
        d1 = (log(Si / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
        lhs = K - Si
        rhs = (
            bs_price(False, Si, K, T, r, b, v)
            - (1 - exp((b - r) * T) * cdf(-d1)) * Si / q1
        )
        bi = (
            -exp((b - r) * T) * cdf(-d1) * (1 - 1 / q1)
            - (1 + exp((b - r) * T) * cdf(-d1) / (v * sqrt(T))) / q1
        )

    return Si


def _put_price(
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:

    Sk = _kp(K, T, r, b, v)
    n = 2 * b / v ** 2
    k = 2 * r / (v ** 2 * (1 - exp(-r * T)))
    d1 = (log(Sk / K) + (b + v ** 2 / 2) * T) / (v * sqrt(T))
    q1 = (-(n - 1) - sqrt((n - 1) ** 2 + 4 * k)) / 2
    a1 = -(Sk / q1) * (1 - exp((b - r) * T) * cdf(-d1))

    if S > Sk:
        return bs_price(False, S, K, T, r, b, v) + a1 * (S / Sk) ** q1
    else:
        return K - S


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    """The Barone-Adesi and Whaley (1987) American approximation.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        b (float): The cost of carry.
        v (float): The asset volatility.

    Returns:
        float: The price of the option.
    """

    if is_call:
        return _call_price(S, K, T, r, b, v)
    else:
        return _put_price(S, K, T, r, b, v)



"""implied volatility"""

from typing import Callable


def solve_ivol(
        p: float,
        price: Callable[[float], float],
        *,
        max_iterations: int = 20,
        epsilon=1e-8
) -> float:

    v_lo = 0.005
    v_hi = 4.0
    p_lo = price(v_lo)
    p_hi = price(v_hi)

    n = 0
    v = v_lo + (p - p_lo) * (v_hi - v_lo) / (p_hi - p_lo)
    p1 = price(v)
    while abs(p - p1) > epsilon and n < max_iterations:
        n += 1

        if p1 < p:
            v_lo = v
        else:
            v_hi = v

        p_lo = price(v_lo)
        p_hi = price(v_hi)
        v = v_lo + (p - p_lo) * (v_hi - v_lo) / (p_hi - p_lo)
        p1 = price(v)

    return v



"""Class for calculating numeric greeks for an option"""

from typing import Callable, Literal

OptionValue = Callable[
    [
        bool,  # is_call
        float,  # Asset price.
        float,  # Strike.
        float,  # Time to expiry in years.
        float,  # Risk free rate.
        float  # Asset volatility
    ],
    float  # The option price
]
DifferenceMethod = Literal['central', 'forward', 'backward']


class NumericGreeks:

    def __init__(
            self,
            price: OptionValue
    ) -> None:
        self.price = price

    def delta(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            method: DifferenceMethod = 'central'
    ) -> float:
        if method == 'central':
            return (
                self.price(is_call, S + dS, K, T, r, v) -
                self.price(is_call, S - dS, K, T, r, v)
            ) / (2 * dS)
        elif method == 'forward':
            return (
                self.price(is_call, S + dS, K, T, r, v) -
                self.price(is_call, S, K, T, r, v)
            ) / dS
        elif method == 'backward':
            return (
                self.price(is_call, S, K, T, r, v) -
                self.price(is_call, S - dS, K, T, r, v)
            ) / dS
        else:
            raise ValueError("Invalid method")

    def gamma(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            method: DifferenceMethod = 'central'
    ) -> float:
        if method == 'central':
            return (
                self.price(is_call, S + dS, K, T, r, v)
                - 2 * self.price(is_call, S, K, T, r, v)
                + self.price(is_call, S - dS, K, T, r, v)
            ) / dS ** 2
        elif method == 'forward':
            return (
                self.price(is_call, S + 2 * dS, K, T, r, v)
                - 2 * self.price(is_call, S + dS, K, T, r, v)
                + self.price(is_call, S, K, T, r, v)
            ) / dS ** 2
        if method == 'central':
            return (
                self.price(is_call, S, K, T, r, v)
                - 2 * self.price(is_call, S - dS, K, T, r, v)
                + self.price(is_call, S - 2 * dS, K, T, r, v)
            ) / dS ** 2
        else:
            raise ValueError("Invalid method")

    def theta(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dT: float = 1 / 365,
            method: DifferenceMethod = 'central'
    ) -> float:
        if method == 'central':
            return (
                self.price(is_call, S, K, T - dT, r, v)
                - self.price(is_call, S, K, T + dT, r, v)
            ) / (2 * dT)
        elif method == 'forward':
            return (
                self.price(is_call, S, K, T, r, v)
                - self.price(is_call, S, K, T + dT, r, v)
            ) / dT
        elif method == 'backward':
            return (
                self.price(is_call, S, K, T - dT, r, v)
                - self.price(is_call, S, K, T, r, v)
            ) / dT
        else:
            raise ValueError("Invalid method")

    def vega(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dv: float = 0.001,
            method: DifferenceMethod = 'central'
    ) -> float:
        if method == 'central':
            return (
                self.price(is_call, S, K, T, r, v + dv)
                - self.price(is_call, S, K, T, r, v - dv)
            ) / (2 * dv)
        elif method == 'forward':
            return (
                self.price(is_call, S, K, T, r, v + dv)
                - self.price(is_call, S, K, T, r, v)
            ) / dv
        elif method == 'backward':
            return (
                self.price(is_call, S, K, T, r, v)
                - self.price(is_call, S, K, T, r, v - dv)
            ) / dv
        else:
            raise ValueError('Invalid method')

    def rho(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dr: float = 0.001,
            method: DifferenceMethod = 'central'
    ) -> float:
        if method == 'central':
            return (
                self.price(is_call, S, K, T, r + dr, v)
                - self.price(is_call, S, K, T, r - dr, v)
            ) / (2 * dr)
        elif method == 'forward':
            return (
                self.price(is_call, S, K, T, r + dr, v)
                - self.price(is_call, S, K, T, r, v)
            ) / dr
        elif method == 'backward':
            return (
                self.price(is_call, S, K, T, r, v)
                - self.price(is_call, S, K, T, r - dr, v)
            ) / dr
        else:
            raise ValueError('Invalid method')

    def elasticity(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            method: DifferenceMethod = 'central'
    ) -> float:
        return (
            self.delta(is_call, S, K, T, r, v, dS=dS, method=method)
            * S
            / self.price(is_call, S, K, T, r, v)
        )

    def speed(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
    ) -> float:
        return (
            self.price(is_call, S + 2 * dS, K, T, r, v)
            - 3 * self.price(is_call, S + dS, K, T, r, v)
            + 3 * self.price(is_call, S, K, T, r, v)
            - self.price(is_call, S - dS, K, T, r, v)
        ) / dS ** 3

    def deltap(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
    ) -> float:
        return (
            self.price(is_call, S * (1 + dS), K, T, r, v) -
            self.price(is_call, S * (1 - dS), K, T, r, v)
        ) / (2 * S)

    def gammap(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            method: DifferenceMethod = 'central'
    ) -> float:
        return S / 100 * self.gamma(is_call, S, K, T, r, v, dS=dS, method=method)

    def vegap(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dv: float = 0.01,
    ) -> float:
        return (
            self.price(is_call, S, K, T, r, v + dv) -
            self.price(is_call, S, K, T, r, v - dv)
        ) * v / 0.1 / 2

    def vanna(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            dv: float = 0.001
    ) -> float:
        # Also known as DdeltaDvol
        return (
            self.price(is_call, S + dS, K, T, r, v + dv) -
            self.price(is_call, S + dS, K, T, r, v - dv) -
            self.price(is_call, S - dS, K, T, r, v + dv) +
            self.price(is_call, S - dS, K, T, r, v - dv)
        ) / (4 * dS) / dv

    def charm(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            dT: float = 1 / 365
    ) -> float:
        # Also known as DdeltaDtime
        return (
            self.price(is_call, S + dS, K, T + dT, r, v) -
            self.price(is_call, S + dS, K, T - dT, r, v) -
            self.price(is_call, S - dS, K, T + dT, r, v) +
            self.price(is_call, S - dS, K, T - dT, r, v)
        ) / (4 * dS) / -dT

    def dgamma_dvol(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dS: float = 0.01,
            dv: float = 0.01
    ) -> float:
        return (
            self.price(is_call, S + dS, K, T, r, v + dv) -
            2 * self.price(is_call, S, K, T, r, v + dv) +
            self.price(is_call, S - dS, K, T, r, v + dv) -
            self.price(is_call, S + dS, K, T, r, v - dv) +
            2 * self.price(is_call, S, K, T, r, v - dv) -
            self.price(is_call, S - dS, K, T, r, v - dv)
        ) / (2 * dv * dS ** 2) / 100

    def vomma(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dv: float = 0.001,
    ) -> float:
        # DvegaDvol
        return (
            self.price(is_call, S, K, T, r, v + dv) -
            2 * self.price(is_call, S, K, T, r, v) +
            self.price(is_call, S, K, T, r, v - dv)
        ) / dv ** 2

    def time_gamma(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dT: float = 1 / 365,
    ) -> float:
        return (
            self.price(is_call, S, K, T + dT, r, v) -
            2 * self.price(is_call, S, K, T, r, v) +
            self.price(is_call, S, K, T - dT, r, v)
        ) / dT ** 2

    def strike_delta(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dX: float = 0.01,
    ) -> float:
        return (
            self.price(is_call, S, K + dX, T, r, v)
            - self.price(is_call, S, K - dX, T, r, v)
        ) / (2 * dX)

    def strike_gamma(
            self,
            is_call: bool,
            S: float,
            K: float,
            T: float,
            r: float,
            v: float,
            *,
            dX: float = 0.01,
    ) -> float:
        return (
            self.price(is_call, S, K + dX, T, r, v)
            - 2 * self.price(is_call, S, K, T, r, v)
            + self.price(is_call, S, K - dX, T, r, v)
        ) / dX ** 2



"""The Bjerksund and Stensland (2002)"""

from math import exp, log, sqrt
from statistics import NormalDist
from typing import Callable

from ..distributions import CBND as cbnd
from ..european.generalised_black_scholes import price as bs_price

norm = NormalDist()
cdf = norm.cdf


def _phi(
        S: float,
        T: float,
        gamma_: float,
        h: float,
        i: float,
        r: float,
        b: float,
        v: float,
) -> float:
    lambda_ = (-r + gamma_ * b + 0.5 * gamma_ * (gamma_ - 1) * v ** 2) * T
    d = -(log(S / h) + (b + (gamma_ - 0.5) * v ** 2) * T) / (v * sqrt(T))
    kappa = 2 * b / v ** 2 + 2 * gamma_ - 1
    return (
        exp(lambda_) * S ** gamma_ * (
            cdf(d)
            - (i / S) ** kappa * cdf(d - 2 * log(i / S) / (v * sqrt(T))))
    )


def _ksi(
        S: float,
        T2: float,
        gamma_: float,
        h: float,
        I2: float,
        I1: float,
        t1: float,
        r: float,
        b: float,
        v: float,
) -> float:
    e1 = (log(S / I1) + (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e2 = (log(I2 ** 2 / (S * I1)) +
          (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e3 = (log(S / I1) - (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))
    e4 = (log(I2 ** 2 / (S * I1)) -
          (b + (gamma_ - 0.5) * v ** 2) * t1) / (v * sqrt(t1))

    f1 = (log(S / h) + (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f2 = (log(I2 ** 2 / (S * h)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f3 = (log(I1 ** 2 / (S * h)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))
    f4 = (log(S * I1 ** 2 / (h * I2 ** 2)) +
          (b + (gamma_ - 0.5) * v ** 2) * T2) / (v * sqrt(T2))

    rho = sqrt(t1 / T2)
    lambda_ = -r + gamma_ * b + 0.5 * gamma_ * (gamma_ - 1) * v ** 2
    kappa = 2 * b / (v ** 2) + (2 * gamma_ - 1)

    return (
        exp(lambda_ * T2) *
        S ** gamma_ *
        (
            cbnd(-e1, -f1, rho) -
            (I2 / S) ** kappa * cbnd(-e2, -f2, rho) -
            (I1 / S) ** kappa * cbnd(-e3, -f3, -rho) +
            (I1 / I2) ** kappa * cbnd(-e4, -f4, -rho)
        )
    )


def _call_price(
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:

    t1 = 1 / 2 * (sqrt(5) - 1) * T

    if b >= r:
        # Use Black-Scholes as it is never optimal to exercise before maturity.
        return bs_price(True, S, K, T, r, b, v)

    beta = (
        (1 / 2 - b / v ** 2)
        + sqrt((b / v ** 2 - 1 / 2) ** 2 + 2 * r / v ** 2)
    )
    b_infinity = beta / (beta - 1) * K
    b0 = max(K, r / (r - b) * K)

    ht1 = -(b * t1 + 2 * v * sqrt(t1)) * K ** 2 / ((b_infinity - b0) * b0)
    ht2 = -(b * T + 2 * v * sqrt(T)) * K ** 2 / ((b_infinity - b0) * b0)
    I1 = b0 + (b_infinity - b0) * (1 - exp(ht1))
    I2 = b0 + (b_infinity - b0) * (1 - exp(ht2))
    alfa1 = (I1 - K) * I1 ** (-beta)
    alfa2 = (I2 - K) * I2 ** (-beta)

    if S >= I2:
        return S - K
    else:
        return (
            alfa2 * S ** beta
            - alfa2 * _phi(S, t1, beta, I2, I2, r, b, v)
            + _phi(S, t1, 1, I2, I2, r, b, v)
            - _phi(S, t1, 1, I1, I2, r, b, v)
            - K * _phi(S, t1, 0, I2, I2, r, b, v)
            + K * _phi(S, t1, 0, I1, I2, r, b, v)
            + alfa1 * _phi(S, t1, beta, I1, I2, r, b, v)
            - alfa1 * _ksi(S, T, beta, I1, I2, I1, t1, r, b, v)
            + _ksi(S, T, 1, I1, I2, I1, t1, r, b, v)
            - _ksi(S, T, 1, K, I2, I1, t1, r, b, v)
            - K * _ksi(S, T, 0, I1, I2, I1, t1, r, b, v)
            + K * _ksi(S, T, 0, K, I2, I1, t1, r, b, v)
        )


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        b: float,
        v: float,
) -> float:
    """The Bjerksund and Stensland (2002) American approximation.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        b (float): The cost of carry of the asset.
        v (float): The volatility of the asset.

    Returns:
        float: The price of the option.
    """

    if is_call:
        return _call_price(S, K, T, r, b, v)
    else:
        # Use the Bjerksund and Stensland put-call transformation
        return _call_price(K, S, T, r - b, -b, v)



"""Black (1976) Options on futures/forwards"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..implied_volatility import solve_ivol

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf


def price(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """Fair value of a futures/forward using Black 76.

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The option price.
    """

    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return exp(-r * T) * (F * cdf(d1) - K * cdf(d2))
    else:
        return exp(-r * T) * (K * cdf(-d2) - F * cdf(-d1))


def ivol(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        p: float,
        *,
        max_iterations: int = 20,
        epsilon = 1e-8
) -> float:
    """Calculate the volatility of a Black 76 option that is implied by the price.

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 20.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, F, K, T, r, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def delta(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """The sensitivity of the option to a change in the asset price
    using Black 76.

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The delta.
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    if is_call:
        return exp(-r * T) * cdf(d1)
    else:
        return -exp(-r * T) * cdf(-d1)


def gamma(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """The second derivative to the change in asset price using Black 76.

    Args:
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The gamma.
    """
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))

    return exp(-r * T) * pdf(d1) / (F * v * sqrt(T))


def theta(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """The change in the value of the option with respect to time to expiry
    using Black 76.

    This value is typically reported by dividing by 365 (for a one calendar day
    movement) or 252 (for a 1 trading day movement).

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The theta.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    if is_call:
        return (
            -F * exp(-r * T) * pdf(d1) * v / (2 * sqrt(T))
            + r * F * exp(-r * T) * cdf(d1)
            - r * K * exp(-r * T) * cdf(d2)
        )
    else:
        return (
            -F * exp(-r * T) * pdf(d1) * v / (2 * sqrt(T))
            - r * F * exp(-r * T) * cdf(-d1)
            + r * K * exp(-r * T) * cdf(-d2)
        )


def vega(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """The sensitivity of the options price or a change in the asset volatility
    using Black 76.

    This value is typically reported by dividing by 100 (for a 1% change in
    volatility)

    Args:
        F (float): The current futures price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The volatility.

    Returns:
        float: The vega.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    return F * exp(-r * T) * pdf(d1) * sqrt(T)


def rho(
        is_call: bool,
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """The sensitivity of the option price to a change in the risk free rate
    using Black 76.

    This value is typically reported by dividing by 100 (for a 1% change
    in the risk free rate)

    Args:
        is_call (bool): True for a call, false for a put.
        F (float): The price of the future.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The rho.
    """
    d1 = (log(F / K) + (v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return -T * exp(-r * T) * (F * cdf(d1) - K * cdf(d2))
    else:
        return -T * exp(-r * T) * (K * cdf(-d2) - F * cdf(-d1))


def vanna(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return -exp(-r * T) * pdf(d1) * d2 / v 


def vomma(
        F: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    d1 = (log(F / K) + T * (v ** 2 / 2)) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return F * exp(-r * T) * pdf(d1) * sqrt(T) * d1 * d2 / v



"""Garman and Kolhagen (1983) Currency options"""

from math import exp, log, sqrt
from statistics import NormalDist

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf
inv_cdf = norm.inv_cdf


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        rf: float,
        v: float,
) -> float:
    """Garman and Kolhagen (1983) Currency options.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate of the base currency.
        rf (float): The risk free rate of the quote currency.
        v (float): The asset volatility.

    Returns:
        float: _description_
    """
    # Garman and Kolhagen (1983) Currency options
                
    d1 = (log(S / K) + (r - rf + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return S * exp(-rf * T) * cdf(d1) - K * exp(-r * T) * cdf(d2)
    else:
        return K * exp(-r * T) * cdf(-d2) - S * exp(-rf * T) * cdf(-d1)



"""Black-Scholes 1973"""

from math import exp, log, sqrt
from statistics import NormalDist

from ..implied_volatility import solve_ivol

norm = NormalDist()
cdf = norm.cdf
pdf = norm.pdf


def price(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    """Black-Scholes for a non-dividend paying stock.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The asset price.
        K (float): The strike price.
        T (float): The time to expiry in years.
        r (float): The risk free rate.
        v (float): The asset volatility.

    Returns:
        float: The price of the option.
    """
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    if is_call:
        return S * cdf(d1) - K * exp(-r*T) * cdf(d2)
    else:
        return K * exp(-r * T) * cdf(-d2) - S * cdf(-d1)


def ivol(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        p: float,
        *,
        max_iterations: int = 35,
        epsilon=1e-8
) -> float:
    """Calculate the volatility of a Black-Scholes 73 option that is implied by
    the price.

    Args:
        is_call (bool): True for a call, false for a put.
        S (float): The current asset price.
        K (float): The option strike price
        T (float): The time to maturity of the option in years.
        r (float): The risk free rate.
        p (float): The option price.
        max_iterations (int, Optional): The maximum number of iterations before
            a price is returned. Defaults to 35.
        epsilon (float, Optional): The largest acceptable error. Defaults to 1e-8.

    Returns:
        float: The implied volatility.
    """
    return solve_ivol(
        p,
        lambda v: price(is_call, S, K, T, r, v),
        max_iterations=max_iterations,
        epsilon=epsilon
    )


def delta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    if is_call:
        return cdf(d1)
    else:
        return -cdf(-d1)

def gamma (
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    "Calculates option gamma"
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    return pdf(d1) / (S * v * sqrt(T))

def theta(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    "Calculates option theta"
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    
    if is_call:
        return - (
            (S * pdf(d1) * v) / (2 * sqrt(T))
        ) - r * K * exp(-r * T) * cdf(d2)

    else:
        return - (
            (S * pdf(d1) * v) / (2 * sqrt(T))
        ) + r * K * exp(-r * T) * cdf(-d2)


def vega (
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    return S * sqrt(T) * pdf(d1)


def rho(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    
    if is_call:
        return K * T * exp(-r * T) * cdf(d2)
    else:
        return -K * T * exp(-r * T) * cdf(-d2)


def vanna(
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    # Also known as DdeltaDvol.
    d1 = (log(S / K) + (r + v * v / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return -d2 * pdf(d1) / v

def charm(
        is_call: bool,
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    # Also known as DdeltaDtime

    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)

    if is_call:
        return -pdf(d1) * (r / (v * sqrt(T)) - d2 / (2 * T))
    else:
        return -pdf(d1) * (r / (v * sqrt(T)) - d2 / (2 * T))


def vomma(
        S: float,
        K: float,
        T: float,
        r: float,
        v: float,
) -> float:
    # Also known as DvegaDvol
    d1 = (log(S / K) + (r + v ** 2 / 2) * T) / (v * sqrt(T))
    d2 = d1 - v * sqrt(T)
    return vega(S, K, T, r, v) * d1 * d2 / v
